package com.company.tab2tables.web.screens.stockmovement;

import com.company.tab2tables.entity.Direction;
import com.haulmont.chile.core.model.MetaClass;
import com.haulmont.chile.core.model.MetaProperty;
import com.haulmont.chile.core.model.MetaPropertyPath;
import com.haulmont.cuba.core.entity.Entity;
import com.haulmont.cuba.core.global.DataManager;
import com.haulmont.cuba.core.global.Messages;
import com.haulmont.cuba.core.global.Metadata;
import com.haulmont.cuba.gui.UiComponents;
import com.haulmont.cuba.gui.components.*;
import com.haulmont.cuba.gui.components.data.table.ContainerGroupTableItems;
import com.haulmont.cuba.gui.model.CollectionContainer;
import com.haulmont.cuba.gui.model.CollectionLoader;
import com.haulmont.cuba.gui.model.impl.CollectionContainerImpl;
import com.haulmont.cuba.gui.screen.*;
import com.company.tab2tables.entity.StockMovement;

import javax.annotation.Nullable;
import javax.inject.Inject;
import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

@UiController("tab2tables_StockMovement.analysis")
@UiDescriptor("stock-movement-analysis.xml")
@LoadDataBeforeShow
public class StockMovementAnalysis extends Screen {

    Map<TabSheet, TabSheet.Tab> tabs = new HashMap<>();
    @Inject
    private TabSheet mainTabSheet;
    @Inject
    private UiComponents uiComponents;
    @Inject
    private Metadata metadata;
    @Inject
    private GroupTable<StockMovement> globalTable;
    @Inject
    private Messages messages;
    @Inject
    private DataManager dataManager;

    @Subscribe
    public void onInit(InitEvent event) {
        if (dataManager.load(StockMovement.class).list().size() == 0) {
            List<StockMovement> tosave = new ArrayList<>();
            int number = 1;
            for (int i = 1; i < 50; i++) {
                StockMovement move = metadata.create(StockMovement.class);
                move.setDate(new Date());
                move.setDirection(Direction.IN);
                move.setStatus("Done");
                move.setPrice("123");
                move.setType("entrée par inventaire");
                move.setWarehouse("entrepôt n° 1");
                move.setSource("client n°12345678");
                move.setDestination("entrepôt");
                move.setQuantity(BigDecimal.valueOf(100));
                move.setUnitsPerPackage(String.valueOf(1000));
                move.setUnitsPerPrice(String.valueOf(1000));
                move.setPrice("11,23");
                move.setNumber(number++);
                tosave.add(move);
            }
            for (int i = 1; i < 50; i++) {
                StockMovement move = metadata.create(StockMovement.class);
                move.setDate(new Date());
                move.setDirection(Direction.IN);
                move.setStatus("Done");
                move.setPrice("123");
                move.setType("entrée par inventaire");
                move.setWarehouse("entrepôt n° 2");
                move.setSource("client n°12345678");
                move.setDestination("entrepôt");
                move.setQuantity(BigDecimal.valueOf(100));
                move.setUnitsPerPackage(String.valueOf(1000));
                move.setUnitsPerPrice(String.valueOf(1000));
                move.setPrice("11,23");
                tosave.add(move);
                move.setNumber(number++);
            }
            for (int i = 1; i < 50; i++) {
                StockMovement move = metadata.create(StockMovement.class);
                move.setDate(new Date());
                move.setDirection(Direction.OUT);
                move.setStatus("Done");
                move.setPrice("123");
                move.setType("entrée par inventaire");
                move.setWarehouse("entrepôt n° 1");
                move.setSource("client n°12345678");
                move.setDestination("entrepôt");
                move.setQuantity(BigDecimal.valueOf(150));
                move.setUnitsPerPackage(String.valueOf(1000));
                move.setUnitsPerPrice(String.valueOf(1000));
                move.setPrice("11,23");
                tosave.add(move);
                move.setNumber(number++);
            }
            for (int i = 1; i < 50; i++) {
                StockMovement move = metadata.create(StockMovement.class);
                move.setDate(new Date());
                move.setDirection(Direction.OUT);
                move.setStatus("Done");
                move.setPrice("123");
                move.setType("entrée par inventaire");
                move.setWarehouse("entrepôt n° 2");
                move.setSource("client n°12345678");
                move.setDestination("entrepôt");
                move.setQuantity(BigDecimal.valueOf(150));
                move.setUnitsPerPackage(String.valueOf(1000));
                move.setUnitsPerPrice(String.valueOf(1000));
                move.setPrice("11,23");
                tosave.add(move);
                move.setNumber(number++);
            }
            dataManager.commit(tosave.toArray(new Entity[0]));
        }
    }

    @Subscribe(id = "stockMovementsDl", target = Target.DATA_LOADER)
    public void onStockMovementsDlPostLoad(CollectionLoader.PostLoadEvent<StockMovement> event) {
        generateTabs(event.getLoadedEntities());
    }

    void generateTabs(List<StockMovement> moves) {
        tabs.forEach((s, t) -> {
            s.removeTab(t.getName());
        });
        List<StockMovement> in = moves.stream().filter(e -> e.getDirection() == Direction.IN).collect(Collectors.toList());
        List<StockMovement> out = moves.stream().filter(e -> e.getDirection() == Direction.OUT).collect(Collectors.toList());
        Map<String, List<StockMovement>> inByWarehouse = in.stream().collect(Collectors.groupingBy(StockMovement::getWarehouse));
        Map<String, List<StockMovement>> outByWarehouse = out.stream().collect(Collectors.groupingBy(StockMovement::getWarehouse));
        Set<String> warehouses = moves.stream().map(StockMovement::getWarehouse)
                .sorted()
                .collect(Collectors.toCollection(LinkedHashSet::new));
        GroupTable<StockMovement> inGlobal = generateTable("Entrées", in);
        GroupTable<StockMovement> outGlobal = generateTable("Sorties", out);
        generateTab(mainTabSheet, "inOutGlobal", "Entrées/Sorties", inGlobal, outGlobal);
        warehouses.forEach(w -> {
            GroupTable<StockMovement> inWarehouse = null;
            GroupTable<StockMovement> outWarehouse = null;
            List<StockMovement> inw = inByWarehouse.get(w);
            if (inw != null)
                inWarehouse = generateTable("Entrées", inw);
            List<StockMovement> outw = outByWarehouse.get(w);
            if (outw != null)
                outWarehouse = generateTable("Sorties", outw);
            generateTab(mainTabSheet, "inOut" + w, w, inWarehouse, outWarehouse);
        });
    }

    TabSheet.Tab generateTab(TabSheet tabSheet, String tabId, String tabCaption, Component... tables) {
        VBoxLayout vbox = uiComponents.create(VBoxLayout.class);
        HBoxLayout hbox = uiComponents.create(HBoxLayout.class);
        hbox.setWidthFull();
        List<Component> comps = Arrays.stream(tables).filter(Objects::nonNull).collect(Collectors.toList());
        float ratio = 1f / comps.size();
        float percent = 100f / comps.size();
        comps.forEach(table -> {
            hbox.add(table);
            //hbox.setExpandRatio(table, ratio);
            //table.setWidth(percent+"%");
        });
        vbox.add(hbox);
        TabSheet.Tab tab = tabSheet.addTab(tabId, vbox);
        tab.setCaption(tabCaption);
        tabs.put(tabSheet, tab);
        return tab;
    }

    GroupTable<StockMovement> generateTable(String tableCaption, List<StockMovement> list) {
        GroupTable<StockMovement> table = createTable(tableCaption);
        MetaClass mc = metadata.getClassNN(StockMovement.class);
        CollectionContainer<StockMovement> dc = new CollectionContainerImpl<>(mc);
        table.setItems(new ContainerGroupTableItems<>(dc));
        new ArrayList<>(table.getColumns()).forEach(table::removeColumn);
        globalTable.getColumns().forEach(column -> {
            MetaPropertyPath mpp = mc.getPropertyPath(column.getIdString());
            String caption = messages.getMessage(StockMovement.class, mpp.toString());
            Table.Column<StockMovement> newCol = new Table.Column<>(mpp, caption);
            table.addColumn(newCol);
            AggregationInfo agg = getAggregationInfo(mpp);
            if (agg != null)
                newCol.setAggregation(agg);
            Table.ColumnAlignment alignment = getAlignment(mpp);
            if (alignment != null)
                newCol.setAlignment(alignment);
        });
        dc.setItems(list);
        return table;
    }

    GroupTable<StockMovement> createTable(String caption) {
        GroupTable<StockMovement> table = uiComponents.create(GroupTable.class);
        table.setMultiSelect(false);
        table.setShowTotalAggregation(true);
        table.setAggregatable(true);
        table.setWidthFull();
        table.setHeightFull();
        table.setSortable(true);
        table.setCaption(caption);
        return table;
    }

    @Nullable
    AggregationInfo getAggregationInfo(MetaPropertyPath mpp) {
        AggregationInfo agg = null;
        if (Set.of("signedQuantity").contains(mpp.toString())) {
            agg = new AggregationInfo();
            agg.setPropertyPath(mpp);
            agg.setType(AggregationInfo.Type.SUM);
        }
        return agg;
    }

    @Nullable
    Table.ColumnAlignment getAlignment(MetaPropertyPath mpp) {
        MetaProperty mp = mpp.getMetaProperty();
        if (mp != null && mp.getRange().isDatatype() && Number.class.isAssignableFrom(mp.getRange().asDatatype().getJavaClass())) {
            return Table.ColumnAlignment.RIGHT;
        }
        return null;
    }

}